package proxy.util;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;

import proxy.Proxy;

public class Log {
	private static enum LogLevel {
		ERROR("ERROR", 31),
		WARN("WARN", 33),
		INFO("INFO", 32),
		DEBUG("DEBUG", 35);
		
		private final String name;
		private final int color;
		
		private LogLevel(String name, int color) {
			this.name = name;
			this.color = color;
		}
	}
	
	private static final boolean DEBUG = System.getProperty("log.debug") != null;
	private static final boolean COLORS = !System.getProperty("os.name").toLowerCase(Locale.US).contains("win") && System.getProperty("log.nocolor") == null;
	private static final DateFormat FORMAT = new SimpleDateFormat("HH:mm:ss");
	
	private static void log(LogLevel level, String name, String msg) {
		System.err.printf(COLORS ? "\u001b[34m[%s]\u001b[m \u001b[" + level.color + "m[%s/%s]\u001b[m \u001b[36m(%s)\u001b[m \u001b[0m%s\n" : "[%s] [%s/%s] (%s) %s\n", 
			FORMAT.format(new Date()), Thread.currentThread().getName(), level.name, name, msg);
	}
	
	private static void log(LogLevel level, String msg) {
		log(level, Proxy.NAME, msg);
	}
	
	public static void error(String msg) {
		log(LogLevel.ERROR, msg);
	}
	
	public static void error(String fmt, Object ... args) {
		log(LogLevel.ERROR, String.format(fmt, args));
	}
	
	public static void error(Throwable t, String msg) {
		log(LogLevel.ERROR, msg);
		StringWriter sw = new StringWriter();
		PrintWriter st = new PrintWriter(sw);
		t.printStackTrace(st);
		sw.flush();
		log(LogLevel.ERROR, sw.toString().trim().replace("\t", "  "));
	}
	
	public static void error(Throwable t, String fmt, Object ... args) {
		log(LogLevel.ERROR, String.format(fmt, args));
		StringWriter sw = new StringWriter();
		PrintWriter st = new PrintWriter(sw);
		t.printStackTrace(st);
		sw.flush();
		log(LogLevel.ERROR, sw.toString().trim().replace("\t", "  "));
	}
	
	public static void warn(String msg) {
		log(LogLevel.WARN, msg);
	}
	
	public static void warn(String fmt, Object ... args) {
		log(LogLevel.WARN, String.format(fmt, args));
	}
	
	public static void infoAs(String name, String msg) {
		log(LogLevel.INFO, name, msg);
	}
	
	public static void info(String msg) {
		log(LogLevel.INFO, msg);
	}
	
	public static void info(String fmt, Object ... args) {
		log(LogLevel.INFO, String.format(fmt, args));
	}
	
	public static void debug(String msg) {
		if(DEBUG)
			log(LogLevel.DEBUG, msg);
	}
	
	public static void debug(String fmt, Object ... args) {
		if(DEBUG)
			log(LogLevel.DEBUG, String.format(fmt, args));
	}
}
